/*
 *    mpp by Russel J. Yount
 *      (C) Copyright 10/17/86 by Carnegie Mellon University
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#ifndef max
#define max(a,b) ((a) < (b) ? (b) : (a))
#endif

#define	maxinputdepth	32
#define	maxlinesize	2048
#define	macrohashsize	2047
#define	macronamesize	128
#define	maxnestdepth	256

struct nest {
  int status;
};

struct file {
  char *name;
  FILE *stream;
  int lineno;
};

struct macro {
  struct macro *next;
  enum macromode {
    rdwr, rdonly
  } mode;
  char *name;
  char *value;
};

static stripcomments = 0;
static showincludes = 0;
static showdefined = 0;
static debugoutput = 0;
static specialchar = '%';
static struct nest nests[maxnestdepth];
static struct nest *nestp = &nests[0];
static struct nest *lastnestp = &nests[maxnestdepth - 1];

static struct file files[maxinputdepth];
static struct file *lastfilep = &files[maxinputdepth - 1];
static struct file *filep = &files[0];

static char inl[maxlinesize];
static char outline[maxlinesize];
static struct macro *macrohashtable[macrohashsize];

static void
error(char *a0, ...)
{
  va_list ap;
  va_start(ap, a0);
  fprintf(stderr, a0, ap);
  va_end(ap);

  putc('\n', stderr);
  exit(1);
}

static void
fileerror(char *a0, ...)
{
  va_list ap;
  va_start(ap, a0);
  fprintf(stderr, "%s; line %d: ", filep->name, filep->lineno);
  fprintf(stderr, a0, ap);
  va_end(ap);

  putc('\n', stderr);
  exit(1);
}

static char *
strsav(char *s)
{
  char *p;

  if ((p = malloc(strlen(s) + 1)) == NULL)
    error("Out of Memory");
  strcpy(p, s);
  return p;
}

static struct macro **
macrolookup(char *name)
{
  register struct macro **mpp, *mp;
  register char *cp;
  register unsigned hv;

  for (cp = name, hv = 0; *cp; hv += *cp++);
  mpp = &macrohashtable[hv % macrohashsize];
  while ((mp = *mpp) && strcmp(mp->name, name))
    mpp = &mp->next;
  return mpp;
}

static
macroundefine(char *name)
{
  register struct macro **mpp, *mp;

  mpp = macrolookup(name);
  if (mp = *mpp) {
    *mpp = mp->next;
    free(mp->value);
    free(mp->name);
    free(mp);
  }
}

static
macrodefine(char *name, char *value, enum macromode mode)
{
  register struct macro **mpp, *mp;

  mpp = macrolookup(name);
  if (mp = *mpp) {
    if (mp->mode == rdonly)
      return;
    free(mp->value);
  } else {
    if ((mp = (struct macro *) malloc(sizeof(struct macro))) == 0)
        error("Out of memory");
    mp->name = strsav(name);
    mp->next = 0;
    *mpp = mp;
  }
  mp->mode = mode;
  mp->value = strsav(value);
}

static int
macroexpand(char *dst, register char *src, int lineno)
{
  char name[macronamesize];
  register char *np;
  char *rp;
  register struct macro *mp;
  int again = 0;

  rp = dst;

  while (*src) {
    if (*src != '$') {
      *dst++ = *src++;
      continue;
    }
    src++;
    if (*src == '$') {
      *dst++ = '$';
      src++;
      continue;
    }
    np = name;
    if (*src == '{' || *src == '(') {
      src++;
      while (*src) {
	if (*src == '$') {
          if (*(src+1) == '{') {
	    /* Write out what we thought was the variable for next pass */
	    src += 2;
	    *dst++='$';
	    *dst++='{';
	    *np = 0;
	    np = name;
	    while (*np) {
	      *dst++ = *np++;
	    }
	    /* Reset name */
	    np = name;
	    /* Flag for another pass */
	    again++;
          }
	}
	if (*src == '}' || *src == ')') {
	  src++;
	  break;
	}
	if (np >= &name[macronamesize]) {
	  fprintf(stderr, "WARNING: Line %d: Macro name truncated to %s\n",
		  lineno, name);
	  while ((*src) && ((*src != '}') || (*src != ')'))) {
	    /* skip to end of the macro */
	    src++;
	  }
	} else
	  *np++ = *src++;
      }
    } else {
      *np++ = *src++;
    }
    *np = 0;
    if (mp = *macrolookup(name)) 
      dst += macroexpand(dst, mp->value, lineno);
  }
  *dst = 0;
  if (again)
    (void) macroexpand(rp, rp, lineno);
  return strlen(rp);
}

static
readline(char *line)
{
  while (filep >= &files[0]) {
    filep->lineno++;
    if (fgets(line, maxlinesize, filep->stream) != NULL)
      return -1;
    if (fclose(filep->stream) == EOF)
      error("Error closing %s", filep->name);
    if (debugoutput && filep != &files[0])
        printf("# End of included file %s\n", filep->name);
    free(filep->name);
    if (filep == &files[0])
      return 0;
    filep--;
  }
  return 0;
}

static
writeline(char *line)
{
  if (!showdefined)
     fputs(line, stdout);
}


static int
directive(char *what, int *status)
{
  char *arg[3], *cp, *tmp1, *tmp2;
  register struct macro **mpp1, **mpp2;
  int n, tmp, itmp1, itmp2;
  int oper = 0, recurse = *status;

  if (*what++ != specialchar && !recurse)
    return nestp->status;
  if (cp = strrchr(what, '\n'))
    *cp = 0;
  for (n = recurse; n < 2; n++) {
    while (*what == ' ' || *what == '\t')
      what++;
    arg[n] = what;
    while (*what != ' ' && *what != '\t' && *what != 0)
      what++;
    if (*what)
      *what++ = 0;
  }
  while (*what == ' ' || *what == '\t')
    what++;
  arg[2] = what;
  if (recurse) goto process;
  if (strcmp(arg[0], "ifdef") == 0) {
    if (nestp == lastnestp)
      fileerror("If Depth overflow");
    if (nestp->status == 2 || nestp->status == 1) {
      nestp++;
      nestp->status = 2;
      return 1;
    }
    nestp++;
    nestp->status = (*macrolookup(arg[1])) ? 0 : 1;
    return 1;
  }
  if (strcmp(arg[0], "ifndef") == 0) {
    if (nestp == lastnestp)
      fileerror("If Depth overflow");
    if (nestp->status == 2 || nestp->status == 1) {
      nestp++;
      nestp->status = 2;
      return 1;
    }
    nestp++;
    nestp->status = (*macrolookup(arg[1])) ? 1 : 0;
    return 1;
  }
  if (strcmp(arg[0], "if") == 0) {
  process:
    if (nestp == lastnestp)
      fileerror("If Depth overflow");
    if (!recurse)
      if (nestp->status == 2 || nestp->status == 1) {
	nestp++;
	nestp->status = 2;
	return 1;
      }
    if ((strcmp(arg[1], "!defined") != 0) && (strcmp(arg[1], "defined") != 0))
      {
	if (strncmp(arg[2], "==", 2) == 0) 
	  oper = 1;
	if (strncmp(arg[2], "!=", 2) == 0) 
	  oper = 2;
	if (strncmp(arg[2], ">", 1) == 0) 
	  oper = 3;
	if (strncmp(arg[2], "<", 1) == 0) 
	  oper = 4;
	if ((strncmp(arg[2], "=", 1) == 0) && (oper == 0))
	  oper = 5;
	if (strncmp(arg[2], ">=", 2) == 0) 
	  oper = 6;
	if (strncmp(arg[2], "<=", 2) == 0) 
	  oper = 7;
	if (strncmp(arg[2], "<>", 2) == 0) 
	  oper = 8;
	if (oper > 0) {
	  while (*what && *what != ' ' && *what != '\t')
	    what++;
	  if (!*what) 
	    fileerror("Comparison to nothing: %s", arg[1]);
	  *what++ = 0;
	  while (*what == ' ' || *what == '\t')
	    what++;
	  arg[2] = what;
	  while (*what && *what != ' ' && *what != '\t')
	    what++;
	  tmp1 = arg[1];
	  tmp2 = arg[2];
	  if (oper > 2) {
	    itmp1 = strtol(tmp1, (char **)NULL, 10);
	    if (itmp1 == ULONG_MAX)
	      fileerror("Math operator parse error on %s, expanded from %s",
			tmp1, arg[1]);
	    itmp2 = strtol(tmp2, (char **)NULL, 10);
	    if (itmp2 == ULONG_MAX)
	      fileerror("Math operator parse error on %s, expanded from %s",
			tmp2, arg[2]);
	    switch (oper) {
	    case 3:
	      *status = (itmp1 > itmp2) ? 0 : 1;
	      break;
	    case 4:
	      *status = (itmp1 < itmp2) ? 0 : 1;
	      break;
	    case 5:
	      *status = (itmp1 == itmp2) ? 0 : 1;
	      break;
	    case 6:
	      *status = (itmp1 >= itmp2) ? 0 : 1;
	      break;
	    case 7:
	      *status = (itmp1 <= itmp2) ? 0 : 1;
	      break;
	    case 8:
	      *status = (itmp1 != itmp2) ? 0 : 1;
	      break;
	    }
	  } else
	    *status = (strncmp(tmp1, tmp2, max(strlen(tmp1), strlen(tmp2)))
		       == 0) ? 0 : 1;
	  if (oper == 2)
	    *status = 1 - *status;
	  if (!recurse) {
	    nestp++;
	    nestp->status = *status;
	  }
	  return 1;
	} else
	  fileerror("Unknown if subdirective %s", arg[1]);
      }
    while (*what && *what != ' ' && *what != '\t')
      what++;
    if (*what) {
      *what++ = 0;
      while (*what == ' ' || *what == '\t')
	what++;
      if (*what == '|' && *(what+1) == '|') {
	oper = 1;
	what += 2;
      }
      if (*what == '&' && *(what+1) == '&') {
	oper = 2;
	what += 2;
      }
      while (*what && *what != ' ' && *what != '\t')
	what++;
    }
    tmp = (*macrolookup(arg[2])) ? 0 : 1;
    if (strcmp(arg[1], "!defined") == 0) 
      tmp = 1 - tmp;
    if (oper) {
      itmp1 = 1;
      directive(what, &itmp1);
    } 
    switch (oper) {
    case 1:
      *status = ((tmp + itmp1) < 2);
      break;
    case 2:
      *status = ((tmp + itmp1) == 0);
      break;
    default:
      *status = tmp;
    }
    if (!recurse) {
      nestp++;
      nestp->status = (1-*status);
    }
    return 1;
  }
  if (strcmp(arg[0], "else") == 0) {
    if (nestp->status == 2)
      return 1;
    if (nestp == &nests[0])
      fileerror("If less else");
    nestp->status = nestp->status ? 0 : 1;
    return 1;
  }
  if (strcmp(arg[0], "endif") == 0) {
    if (nestp == &nests[0])
      fileerror("If less endif");
    nestp--;
    return 1;
  }
  if (nestp->status)
    return 1;
  if (strcmp(arg[0], "include") == 0) {
    if (filep == lastfilep)
      fileerror("Include file overflow");
    filep++;
    if (showincludes)
      fprintf(stderr, "Reading %s\n", arg[1]);
    if ((filep->stream = fopen(arg[1], "r")) == NULL) {
      filep--;
      if (showdefined) return 1;
      fileerror("Can't open %s", arg[1]);
    }
    filep->name = strsav(arg[1]);
    filep->lineno = 0;
    if (debugoutput)
      printf("# Start of included file %s\n", filep->name);
    return 1;
  }
  if (strcmp(arg[0], "define") == 0) {
    macrodefine(arg[1], arg[2], rdwr);
    if (showdefined) printf("%s %s\n", arg[1], arg[2]);
    return 1;
  }
  if (strcmp(arg[0], "undef") == 0) {
    macroundefine(arg[1]);
    return 1;
  }
  fileerror("Unknown directive %s", arg[0]);
}

void
expandfile(char *name)
{
  int crap = 0;

  if (strcmp(name, "-") == 0) {
    filep->stream = stdin;
    filep->name = strsav("(stdin)");
  } else {
    filep->name = strsav(name);
    if ((filep->stream = fopen(name, "r")) == NULL) {
      fileerror("Can't open %s", name);
      exit(1);
    }
  }
  if (showincludes)
    fprintf(stderr, "Reading %s\n", filep->name);
  filep->lineno = 0;
  while (readline(inl)) {
    if (stripcomments) {
      char *cp;
      for (cp = inl; *cp != 0 && *cp != '#'; cp++)
	continue;
      *cp = 0;
      if (cp == inl)
	continue;
    }
    (void) macroexpand(outline, inl, filep->lineno);
    crap = 0;
    if (directive(outline, &crap)) {
      if (debugoutput) {
        if(inl && *inl)
          if (nestp->status)
            printf("#X");
          else
            printf("#");
        writeline(inl);
      }
      continue;
    }
    writeline(outline);
  }
}

static
usage()
{
  fprintf(stderr,
	  "Usage: mpp [-cC][-s][-v][-d][-m][-Dname=value][-Uname][-][files]\n");
  exit(1);
}

int
main(int argc, char **argv)
{
  argv++, argc--;
  if (argc == 0)
    usage();
  while (argc > 0) {
    if (**argv == '-') {
      if (strcmp(*argv, "-s") == 0) {
	stripcomments++;
	argv++, argc--;
	continue;
      }
      if (strcmp(*argv, "-d") == 0) {
	debugoutput++;
	argv++, argc--;
	continue;
      }
      if (strcmp(*argv, "-m") == 0) {
	showdefined++;
	argv++, argc--;
	continue;
      }
      if (strcmp(*argv, "-v") == 0) {
	showincludes++;
	argv++, argc--;
	continue;
      }
      if (strncmp(*argv, "-c", sizeof("-c") - 1) == 0) {
	specialchar = argv[0][sizeof("-c")];
	argv++, argc--;
	continue;
      }
      if (strncmp(*argv, "-U", sizeof("-U") - 1) == 0) {
	macroundefine(&argv[0][sizeof("-U")]);
	argv++, argc--;
	continue;
      }
      if (strncmp(*argv, "-D", sizeof("-D") - 1) == 0) {
	char *cp, *cp2;

	cp = &argv[0][sizeof("-D") - 1];
	if (cp2 = strrchr(cp, '='))
	  *cp2++ = 0;
	if (cp2 == 0)
	  cp2 = "";
	macrodefine(cp, cp2, rdonly);
	argv++, argc--;
	continue;
      }
      if (strcmp(*argv, "-"))
	usage();
    }
    expandfile(*argv);
    argv++, argc--;
  }
  exit(0);
}
